unit SortingClosure;
// Rob Kennedy
// 4 June 2005
// http://www.cs.wisc.edu/~rkennedy/

// The code in this unit allows for the use of methods as the callback
// camparison function to the TList.Sort method instead of the standalone
// function it usually requires. Call MakeSortInstance on the method you want to
// use, and then pass the result to TList.Sort. Call FreeSortInstance when you
// are finished with the method.

// This code is based on Borland's implementation of MakeObjectInstance in
// Classes.pas.

interface

uses
  Classes;

type
  TSortListMethod = function(Item1, Item2: Pointer): Integer of object;

function MakeSortCompare(Method: TSortListMethod): TListSortCompare;
procedure FreeSortCompare(SortProc: TListSortCompare);

implementation

uses
  Windows, SysUtils;

var
  MinReserveSize: DWord;
  PageSize: DWord;
  MaxPagesPerBlock: Cardinal;
  ProcsPerPage: Cardinal;
  Lock: TRTLCriticalSection;

type
  PSortListMethod = ^TSortListMethod;

  TCallJmpInstruction = packed record
    Opcode: Byte;
    Offset: LongInt;
  end;

  PSortProc = ^TSortProc;
  TSortProc = packed record
    Code: TCallJmpInstruction;
  case Boolean of
    False: (Next: PSortProc);
    True: (
      Method: TSortListMethod;
      // This padding is here in the interest of making the
      // code in this structure be aligned on a 16-byte
      // bounary, which should allow it to run a little better.
      Padding: array[1..3] of Byte;
    );
  end;

  TPageCode = packed record
    PopEcx: Byte;
    Jmp: TCallJmpInstruction;
  end;

  PPage = ^TPage;
  TPage = packed record
    Code: TPageCode;
    Padding: array[1..2] of Byte;
    PreviousPage: PPage;
    PageIndex: Cardinal;
    Procs: array[0..0] of TSortProc;
  end;

procedure InitGlobals;
var
  SystemInfo: TSystemInfo;
begin
  GetSystemInfo(SystemInfo);
  MinReserveSize := SystemInfo.dwAllocationGranularity;
  PageSize := SystemInfo.dwPageSize;
  MaxPagesPerBlock := MinReserveSize div PageSize;
  Assert(MinReserveSize mod PageSize = 0);
  ProcsPerPage := Succ((PageSize - SizeOf(TPage)) div SizeOf(TSortProc));
end;

function StdSortProc(Item1, Item2: Pointer; Method: PSortListMethod): Integer;
begin
  Result := Method^(Item1, Item2);
end;

function CalcJmpOffset(Src, Dest: Pointer): Longint;
begin
  Result := Longint(Dest) - (Longint(Src) + SizeOf(TCallJmpInstruction));
end;

var
  FirstFreeProc: PSortProc = nil;
  MostRecentPage: PPage = nil;

function MakeSortCompare(Method: TSortListMethod): TListSortCompare;
var
  Proc: PSortProc;
  NewPage: PPage;
  i: Cardinal;
begin
  if IsMultithread then EnterCriticalSection(Lock);
  try
    if not Assigned(FirstFreeProc) then begin
      // No unused sort procedures remain. Allocate another page of them.
      if not Assigned(MostRecentPage) then InitGlobals;
      if not Assigned(MostRecentPage) or (MostRecentPage.PageIndex >= Pred(MaxPagesPerBlock)) then begin
        // Nothing has ever been allocated, or we need to allocate a whole new block
        NewPage := VirtualAlloc(nil, MinReserveSize, Mem_Reserve, Page_NoAccess);
        Win32Check(Assigned(NewPage));
        NewPage := VirtualAlloc(NewPage, PageSize, Mem_Commit, Page_Execute_ReadWrite);
        Win32Check(Assigned(NewPage));
        NewPage.PageIndex := 0;
      end else begin
        // Get the next page from the reserved space
        NewPage := VirtualAlloc(Ptr(LongWord(MostRecentPage) + PageSize), PageSize, Mem_Commit, Page_Execute_ReadWrite);
        Win32Check(Assigned(NewPage));
        NewPage.PageIndex := Succ(MostRecentPage.PageIndex);
      end;
      NewPage.PreviousPage := MostRecentPage;

      // Control reaches this Code value via one of the Proc items, which has
      // placed on the stack a pointer to the method that should be called. This
      // code then pops that pointer off the stack and into ECX before jumping
      // to the StdSortProc function. ECX holds the third parameter of that
      // function. The first two were put there by the TList.QuickSort procedure
      // when it called the SCompare function pointer.
      NewPage.Code.PopEcx := $59; // pop ecx
      NewPage.Code.Jmp.Opcode := $e9; // jmp
      NewPage.Code.Jmp.Offset := CalcJmpOffset(@NewPage.Code.Jmp, @StdSortProc);
      MostRecentPage := NewPage;
      Proc := @NewPage.Procs[0];
      for i := 1 to ProcsPerPage do begin
        // The CALL instruction does two things, both of which are important in
        // this situation. It jumps the specified number of bytes, which is a
        // negative number, in this case, to the instruction in NewPage.Code.
        // But before it jumps there, it pushes the address of the NEXT
        // instruction onto the stack. In this case, there is no next
        // instruction. Instead, at that address is the Proc.Method value.
        Proc.Code.Opcode := $e8; // call near ptr Offset
        Proc.Code.Offset := CalcJmpOffset(@Proc.Code, @NewPage.Code);
        Proc.Next := FirstFreeProc;
        FirstFreeProc := Proc;
        Inc(Proc);
      end;
    end;
    Proc := FirstFreeProc;
    FirstFreeProc := Proc.Next;
  finally
    if IsMultithread then LeaveCriticalSection(Lock);
  end;
  Proc.Method := Method;
  Pointer(Result) := Proc;
end;

procedure FreeSortCompare(SortProc: TListSortCompare);
begin
  Assert(Assigned(SortProc));
  if IsMultithread then EnterCriticalSection(Lock);
  try
    PSortProc(@SortProc).Next := FirstFreeProc;
    FirstFreeProc := @SortProc;
  finally
    if IsMultithread then LeaveCriticalSection(Lock);
  end;
end;

function GetPageReservationBase(Page: PPage): PPage;
begin
  Result := Ptr(LongWord(Page) - (LongWord(Page) mod MinReserveSize));
end;

procedure FreeSortProcBlocks;
var
  Page: PPage;
begin
  MostRecentPage := GetPageReservationBase(MostRecentPage);
  while Assigned(MostRecentPage) do begin
    Page := MostRecentPage.PreviousPage;
    VirtualFree(MostRecentPage, 0, Mem_Release);
    MostRecentPage := GetPageReservationBase(Page);
  end;
end;

initialization
  InitializeCriticalSection(Lock);
finalization
  FreeSortProcBlocks;
  DeleteCriticalSection(Lock);
end.
